import shlex
from unittest import mock

import httpx
import pytest

from jobbergate_cli.schemas import JobSubmissionResponse
from jobbergate_cli.subapps.job_submissions.app import (
    HIDDEN_FIELDS,
    cancel,
    clone,
    create,
    delete,
    get_one,
    list_all,
    style_mapper,
)
from jobbergate_cli.text_tools import unwrap


@pytest.mark.parametrize(
    "flag_name,flag_job_script_id,separator",
    [
        ("--name", "--job-script-id", "="),
        ("-n", "-i", " "),
    ],
)
def test_create(
    make_test_app,
    dummy_context,
    dummy_job_submission_data,
    cli_runner,
    mocker,
    flag_name,
    flag_job_script_id,
    separator,
):
    job_submission_data = JobSubmissionResponse(**dummy_job_submission_data[0])
    job_submission_name = job_submission_data.name
    job_submission_description = job_submission_data.description
    job_script_id = job_submission_data.job_script_id

    mocked_render = mocker.patch("jobbergate_cli.subapps.job_submissions.app.render_single_result")

    submissions_handler = mock.MagicMock()
    submissions_handler.run.return_value = job_submission_data
    mocked_factory = mocker.patch(
        "jobbergate_cli.subapps.job_submissions.app.job_submissions_factory", return_value=submissions_handler
    )

    test_app = make_test_app("create", create)
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                create {flag_name}{separator}{job_submission_name}
                       --description='{job_submission_description}'
                       {flag_job_script_id}{separator}{job_script_id}
                       --download
                """
            )
        ),
    )
    assert result.exit_code == 0, f"create failed: {result.stdout}"

    mocked_factory.assert_called_once_with(
        dummy_context,
        job_script_id,
        job_submission_name,
        description=job_submission_description,
        execution_directory=None,
        cluster_name=None,
        sbatch_arguments=None,
        download=True,
    )

    mocked_render.assert_called_once_with(
        dummy_context,
        job_submission_data,
        title="Created Job Submission",
        hidden_fields=HIDDEN_FIELDS,
    )


def test_list_all__renders_paginated_results(
    make_test_app,
    dummy_context,
    cli_runner,
    mocker,
    attach_persona,
):
    test_app = make_test_app("list-all", list_all)
    mocked_pagination = mocker.patch("jobbergate_cli.subapps.job_submissions.app.handle_pagination")
    attach_persona("dummy@dummy.com")
    result = cli_runner.invoke(test_app, ["list-all"])
    assert result.exit_code == 0, f"list-all failed: {result.stdout}"
    mocked_pagination.assert_called_once_with(
        jg_ctx=dummy_context,
        url_path="/jobbergate/job-submissions",
        abort_message="Couldn't retrieve job submissions list from API",
        params={"user_only": True, "sort_ascending": False, "sort_field": "id", "include_archived": False},
        title="Job Submission List",
        style_mapper=style_mapper,
        hidden_fields=HIDDEN_FIELDS,
        nested_response_model_cls=JobSubmissionResponse,
        value_mappers=None,
    )


@pytest.mark.parametrize(
    "flag_id,separator",
    [
        ("--id", "="),
        ("-i", " "),
    ],
)
def test_get_one__success(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_job_submission_data,
    dummy_domain,
    cli_runner,
    mocker,
    flag_id,
    separator,
    attach_persona,
):
    attach_persona("dummy@dummy.com")
    respx_mock.get(f"{dummy_domain}/jobbergate/job-submissions/1").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_job_submission_data[0],
        ),
    )
    test_app = make_test_app("get-one", get_one)
    mocked_render = mocker.patch("jobbergate_cli.subapps.job_submissions.app.render_single_result")
    result = cli_runner.invoke(test_app, shlex.split(f"get-one {flag_id}{separator}1"))
    assert result.exit_code == 0, f"get-one failed: {result.stdout}"
    mocked_render.assert_called_once_with(
        dummy_context,
        JobSubmissionResponse(**dummy_job_submission_data[0]),
        title="Job Submission",
        hidden_fields=HIDDEN_FIELDS,
        value_mappers=None,
    )


@pytest.mark.parametrize(
    "flag_id,separator",
    [
        ("--id", "="),
        ("-i", " "),
    ],
)
def test_delete__makes_request_and_sends_terminal_message(
    respx_mock,
    make_test_app,
    dummy_domain,
    cli_runner,
    flag_id,
    separator,
):
    job_submission_id = 13

    delete_route = respx_mock.delete(f"{dummy_domain}/jobbergate/job-submissions/{job_submission_id}").mock(
        return_value=httpx.Response(httpx.codes.NO_CONTENT),
    )
    test_app = make_test_app("delete", delete)
    result = cli_runner.invoke(test_app, shlex.split(f"delete {flag_id}{separator}{job_submission_id}"))
    assert result.exit_code == 0, f"delete failed: {result.stdout}"
    assert delete_route.called
    assert "JOB SUBMISSION DELETE SUCCEEDED"


class TestCloneJobSubmission:
    @pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
    def test_clone__success(
        self,
        respx_mock,
        make_test_app,
        dummy_job_submission_data,
        dummy_domain,
        dummy_context,
        cli_runner,
        mocker,
        selector_template,
    ):
        """
        Test that the clone application subcommand works as expected.
        """

        job_submission_data = dummy_job_submission_data[0]
        id = job_submission_data["id"]

        cli_selector = selector_template.format(id=id)

        clone_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-submissions/clone/{id}").mock(
            return_value=httpx.Response(
                httpx.codes.CREATED,
                json=job_submission_data,
            ),
        )
        mocked_render = mocker.patch("jobbergate_cli.subapps.job_submissions.app.render_single_result")

        test_app = make_test_app("clone", clone)
        result = cli_runner.invoke(test_app, shlex.split(f"clone {cli_selector}"))

        assert clone_route.called

        assert result.exit_code == 0, f"clone failed: {result.stdout}"
        mocked_render.assert_called_once_with(
            dummy_context,
            JobSubmissionResponse(**job_submission_data),
            title="Cloned Job Submission",
            hidden_fields=HIDDEN_FIELDS,
        )


class TestCancelJobSubmission:
    @pytest.mark.parametrize("selector_template", ["{id}", "-i {id}", "--id={id}", "--id {id}"])
    def test_cancel__success(
        self,
        respx_mock,
        make_test_app,
        dummy_job_submission_data,
        dummy_domain,
        dummy_context,
        cli_runner,
        mocker,
        selector_template,
    ):
        """
        Test that the cancel application subcommand works as expected.
        """

        job_submission_data = dummy_job_submission_data[0]
        id = job_submission_data["id"]

        cli_selector = selector_template.format(id=id)

        cancel_route = respx_mock.put(f"{dummy_domain}/jobbergate/job-submissions/cancel/{id}").mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                json=job_submission_data,
            ),
        )
        mocked_render = mocker.patch("jobbergate_cli.subapps.job_submissions.app.render_single_result")

        test_app = make_test_app("cancel", cancel)
        result = cli_runner.invoke(test_app, shlex.split(f"cancel {cli_selector}"))

        assert cancel_route.called

        assert result.exit_code == 0, f"cancel failed: {result.stdout}"
        mocked_render.assert_called_once_with(
            dummy_context,
            JobSubmissionResponse(**job_submission_data),
            title="Cancelled Job Submission",
            hidden_fields=HIDDEN_FIELDS,
        )



from typing import Any, Dict, Optional, TypeVar

from pydantic import BaseModel

from jobbergate_cli.schemas import ListResponseEnvelope


EnvelopeT = TypeVar("EnvelopeT")


def test_list_response_envelope_with_plain_dict():
    data = dict(
        items=[
            dict(name="item-1"),
            dict(name="item-2"),
        ],
        total=2,
        page=0,
        size=5,
        pages=1,
    )
    envelope: ListResponseEnvelope[Dict[str, Any]] = ListResponseEnvelope[Dict[str, Any]](**data)
    assert envelope.items == [
        dict(name="item-1"),
        dict(name="item-2"),
    ]
    assert envelope.total == 2
    assert envelope.page == 0
    assert envelope.size == 5
    assert envelope.pages == 1


def test_list_envelope_with_nested_model():
    data = dict(
        items=[
            dict(name="item-1"),
            dict(name="item-2"),
        ],
        total=2,
        page=0,
        size=5,
        pages=1,
    )

    class DummyModel(BaseModel):
        name: str

    dummy_envelope: ListResponseEnvelope[DummyModel] = ListResponseEnvelope[DummyModel](**data)
    assert dummy_envelope.items == [
        DummyModel(name="item-1"),
        DummyModel(name="item-2"),
    ]
    assert dummy_envelope.total == 2
    assert dummy_envelope.page == 0
    assert dummy_envelope.size == 5
    assert dummy_envelope.pages == 1

    class IdiotModel(BaseModel):
        name: str
        foo: Optional[str] = None

    idiot_envelope: ListResponseEnvelope[IdiotModel] = ListResponseEnvelope[IdiotModel](**data)
    assert idiot_envelope.items == [
        IdiotModel(name="item-1"),
        IdiotModel(name="item-2"),
    ]
    assert idiot_envelope.total == 2
    assert idiot_envelope.page == 0
    assert idiot_envelope.size == 5
    assert idiot_envelope.pages == 1



import contextlib
import pathlib

import pytest
from loguru import logger

from jobbergate_cli.config import settings


@pytest.fixture
def caplog(caplog):
    handler_id = logger.add(caplog.handler, format="{message}")
    yield caplog
    logger.remove(handler_id)


@pytest.fixture(autouse=True)
def clear_cache():
    """
    Remove any data left in the test cache.

    Data should not have been stored in the cache during tests, but sometimes mistakes are made.
    """

    def deltree(tree: pathlib.Path):
        for item in tree.iterdir():
            if item.is_file() or item.is_symlink():
                item.unlink()
            elif item.is_dir():
                deltree(item)

    deltree(settings.JOBBERGATE_CACHE_DIR)


@pytest.fixture
def tweak_settings():
    """
    Provides a fixture to use as a context manager where the cli settings may be temporarily changed.
    """

    @contextlib.contextmanager
    def _helper(**kwargs):
        """
        Context manager for tweaking app settings temporarily.
        """
        previous_values = {}
        for key, value in kwargs.items():
            previous_values[key] = getattr(settings, key)
            setattr(settings, key, value)
        try:
            yield
        finally:
            for key, value in previous_values.items():
                setattr(settings, key, value)

    return _helper



"""
Provides a data object for context passed from the main entry point.
"""

from dataclasses import dataclass
from functools import cached_property

from buzz import check_expressions
from httpx import Client
from jobbergate_core.auth.handler import JobbergateAuthHandler
from jobbergate_core.sdk import Apps

from jobbergate_cli.auth import show_login_message, track_login_progress
from jobbergate_cli.exceptions import Abort
from jobbergate_cli.config import settings
from jobbergate_cli.schemas import ContextProtocol


@dataclass
class JobbergateContext(ContextProtocol):
    """
    A data object describing context passed from the main entry point.
    """

    raw_output: bool = False
    full_output: bool = False

    @cached_property
    def client(self) -> Client:
        """
        Client for making requests to the Jobbergate API.
        """
        return Client(
            base_url=settings.BASE_API_URL,
            auth=self.authentication_handler,
            timeout=settings.JOBBERGATE_REQUESTS_TIMEOUT,
        )

    @cached_property
    def authentication_handler(self) -> JobbergateAuthHandler:
        """
        The authentication handler for the context.

        This is a cached property to ensure that the handler is only created when needed,
        so commands that require no authentication face no configuration errors.
        """
        with check_expressions(
            "The following settings are required to enable authenticated requests:",
            raise_exc_class=Abort,
            raise_kwargs=dict(support=True, subject="Configuration error"),
        ) as check:
            check(settings.OIDC_DOMAIN, "OIDC_DOMAIN")
            check(settings.OIDC_CLIENT_ID, "OIDC_CLIENT_ID")

        protocol = "https" if settings.OIDC_USE_HTTPS else "http"
        return JobbergateAuthHandler(
            cache_directory=settings.JOBBERGATE_USER_TOKEN_DIR,
            login_domain=f"{protocol}://{settings.OIDC_DOMAIN}",
            login_client_id=settings.OIDC_CLIENT_ID,
            login_client_secret=settings.OIDC_CLIENT_SECRET,
            login_url_handler=show_login_message,
            login_sequence_handler=track_login_progress,
        )

    @cached_property
    def sdk(self) -> Apps:
        """
        SDK for accessing Jobbergate API.
        """
        return Apps(client=self.client)



"""
Configuration file, sets all the necessary environment variables.
Can load configuration from a dotenv file if supplied.
"""

from pathlib import Path
from sys import exit
from typing import Optional

from pydantic import Field, ValidationError, computed_field, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from jobbergate_cli import constants
from jobbergate_cli.constants import OV_CONTACT
from jobbergate_cli.render import terminal_message
from jobbergate_cli.text_tools import conjoin


def _get_env_file() -> Path:
    """
    Load the env file based on the constant JOBBERGATE_DEFAULT_DOTENV_PATH if the file exists. Otherwise use ".env".
    """
    env_file: Path
    if constants.JOBBERGATE_DEFAULT_DOTENV_PATH.is_file():
        env_file = constants.JOBBERGATE_DEFAULT_DOTENV_PATH
    else:
        env_file = Path(".env")
    return env_file


class Settings(BaseSettings):
    """
    Provide a ``pydantic`` settings model to hold configuration values loaded from the environment.
    """

    JOBBERGATE_CACHE_DIR: Path = Field(Path.home() / ".local/share/jobbergate3")

    BASE_API_URL: str = Field("https://apis.vantagecompute.ai")
    ARMADA_API_BASE: Optional[str] = None

    SBATCH_PATH: Optional[Path] = None

    # enable http tracing
    JOBBERGATE_DEBUG: bool = Field(False)
    JOBBERGATE_REQUESTS_TIMEOUT: Optional[int] = 15

    # Setry's configuration
    SENTRY_DSN: Optional[str] = None
    SENTRY_TRACE_RATE: float = Field(1.0, gt=0.0, le=1.0)
    SENTRY_ENV: str = "LOCAL"

    # Default job submission cluster
    DEFAULT_CLUSTER_NAME: Optional[str] = None

    # How long it will use cached cluster lists before fetching them again
    JOBBERGATE_CLUSTER_CACHE_LIFETIME: int = 60 * 5

    # Settings for log uploads
    JOBBERGATE_AWS_ACCESS_KEY_ID: Optional[str] = None
    JOBBERGATE_AWS_SECRET_ACCESS_KEY: Optional[str] = None
    JOBBERGATE_S3_LOG_BUCKET: str = Field("jobbergate-cli-logs")

    # Compatibility mode: If True, add commands as they appear in the legacy app
    JOBBERGATE_COMPATIBILITY_MODE: Optional[bool] = False
    JOBBERGATE_LEGACY_NAME_CONVENTION: Optional[bool] = False

    # Auth0 config for machine-to-machine security
    OIDC_DOMAIN: str = "auth.vantagecompute.ai/realms/vantage"
    OIDC_CLIENT_ID: str = "default"
    OIDC_USE_HTTPS: bool = True
    OIDC_CLIENT_SECRET: Optional[str] = None

    @model_validator(mode="after")
    def _handle_api_base_url_compatibility(self) -> "Settings":
        """
        For backward compatibility, ARMADA_API_BASE takes priority over BASE_API_URL if provided.
        """
        if self.ARMADA_API_BASE:
            self.BASE_API_URL = self.ARMADA_API_BASE
        return self

    @field_validator("JOBBERGATE_CACHE_DIR", mode="after")
    def _validate_cache_dir(cls, value: Path) -> Path:
        """
        Expand, resolve, and create cache directory.
        """
        value = value.expanduser().resolve()
        value.mkdir(exist_ok=True, parents=True)
        return value

    @computed_field
    def JOBBERGATE_USER_TOKEN_DIR(self) -> Path:
        token_dir = self.JOBBERGATE_CACHE_DIR / "token"
        token_dir.mkdir(exist_ok=True, parents=True)
        return token_dir

    @computed_field
    def JOBBERGATE_LOG_PATH(self) -> Path:
        log_file = self.JOBBERGATE_CACHE_DIR / "logs" / "jobbergate-cli.log"
        log_file.parent.mkdir(exist_ok=True, parents=True)
        return log_file

    @computed_field
    def JOBBERGATE_APPLICATION_MODULE_PATH(self) -> Path:
        return self.JOBBERGATE_CACHE_DIR / constants.JOBBERGATE_APPLICATION_MODULE_FILE_NAME

    @computed_field
    def JOBBERGATE_APPLICATION_CONFIG_PATH(self) -> Path:
        return self.JOBBERGATE_CACHE_DIR / constants.JOBBERGATE_APPLICATION_CONFIG_FILE_NAME

    @computed_field
    def JOBBERGATE_CLUSTER_LIST_PATH(self) -> Path:
        return self.JOBBERGATE_CACHE_DIR / "clusters.json"

    @property
    def is_onsite_mode(self) -> bool:
        """Check if the SBATCH_PATH is set, indicating that the CLI is running in on-site mode."""
        return self.SBATCH_PATH is not None

    model_config = SettingsConfigDict(env_file=_get_env_file(), extra="ignore")


def build_settings(*args, **kwargs):
    """
    Return a Setting object and handle ValidationError with a message to the user.
    """
    try:
        return Settings(*args, **kwargs)
    except ValidationError as err:
        terminal_message(
            conjoin(
                "A configuration error was detected.",
                "",
                f"[yellow]Please contact [bold]{OV_CONTACT}[/bold] for support and trouble-shooting[/yellow]",
                "",
                "Details:",
                "",
                f"[red]{err}[/red]",
            ),
            subject="Configuration Error",
        )
        exit(1)


settings = build_settings()



from pathlib import Path

import pydantic
import pytest

from jobbergate_cli.config import Settings


def test_SENTRY_TRACE_RATE__requires_float_in_valid_range():
    assert Settings(SENTRY_TRACE_RATE=0.5)

    with pytest.raises(pydantic.ValidationError):
        Settings(SENTRY_TRACE_RATE=0.0)

    with pytest.raises(pydantic.ValidationError):
        Settings(SENTRY_TRACE_RATE=1.1)


def test_cache_dir__expands_user_and_resolves():
    settings = Settings(JOBBERGATE_CACHE_DIR="~/.jobbergate-cli-prod")
    assert settings.JOBBERGATE_CACHE_DIR == Path.home() / ".jobbergate-cli-prod"
    assert settings.JOBBERGATE_CACHE_DIR.is_absolute()


@pytest.mark.parametrize("sbatch_path, is_onsite_mode", [[None, False], ["/usr/bin/sbatch", True]])
def test_is_onsite_mode__is_true_when_sbatch_path_is_set(sbatch_path, is_onsite_mode):
    settings = Settings(SBATCH_PATH=sbatch_path)
    assert settings.is_onsite_mode == is_onsite_mode


def test_armada_api_base__overrides_base_api_url_when_both_are_set():
    settings = Settings(BASE_API_URL="https://new-url.example.com", ARMADA_API_BASE="https://legacy-url.example.com")
    assert settings.BASE_API_URL == "https://legacy-url.example.com"


def test_armada_api_base__sets_base_api_url_when_only_armada_api_base_is_set():
    settings = Settings(ARMADA_API_BASE="https://legacy-url.example.com")
    assert settings.BASE_API_URL == "https://legacy-url.example.com"



"""
Utilities for handling auth in jobbergate-cli.
"""

import webbrowser
from typing import Iterable

from jobbergate_core.auth.handler import DeviceCodeData
from loguru import logger
from rich.console import Console
from rich.progress import track

from jobbergate_cli.render import terminal_message
from jobbergate_cli.text_tools import copy_to_clipboard


def open_on_browser(url: str) -> bool:
    """Open the url on the browser using webbrowser."""
    try:
        browser = webbrowser.get()
        if isinstance(browser, webbrowser.GenericBrowser):
            # skip all browsers started with a command and without remote functionality
            return False
        return browser.open(url)
    except Exception as e:
        logger.warning(f"Couldn't open login url on browser due to -- {str(e)}")
        return False


def show_login_message(device_code_data: DeviceCodeData):
    """Show a message to the user with a link to the auth provider to login."""
    console = Console()
    EXTRA_CHARS = 7  # for indentation and panel borders

    verification_uri = device_code_data.verification_uri_complete
    waiting_time = int(device_code_data.expires_in / 60)

    kwargs = {}

    if open_on_browser(verification_uri):
        kwargs["footer"] = "The output was opened on your browser"
    elif copy_to_clipboard(verification_uri):
        kwargs["footer"] = "The output was copied to your clipboard"

    if console.width >= len(verification_uri) + EXTRA_CHARS:
        _show_login_standard_screen(verification_uri, waiting_time, **kwargs)
    else:
        _show_login_narrow_screen(verification_uri, waiting_time, console, **kwargs)


def _show_login_narrow_screen(verification_uri: str, waiting_time: int, console: Console, **kwargs):
    """Print the link out of the panel to make it easier to copy."""
    terminal_message(
        f"""
        To complete login, please open the link bellow in a browser.

        Waiting up to {waiting_time} minutes for you to complete the process...
        """,
        subject="Waiting for login",
        **kwargs,
    )
    console.print(verification_uri, overflow="ignore", no_wrap=True, crop=False)
    console.print()


def _show_login_standard_screen(verification_uri: str, waiting_time: int, **kwargs):
    """Print a rich panel with a link to the auth provider to login."""
    terminal_message(
        f"""
        To complete login, please open the following link in a browser:

          {verification_uri}

        Waiting up to {waiting_time} minutes for you to complete the process...
        """,
        subject="Waiting for login",
        **kwargs,
    )


def track_login_progress(iterable: Iterable) -> Iterable:
    """Track the progress of the login process on a progress bar."""
    return track(iterable, description="[green]Waiting for web login...", update_period=1, transient=True)



from jobbergate_cli.subapps.clusters.app import list_all


def test_list_all__makes_request_and_renders_results(
    make_test_app,
    attach_persona,
    seed_clusters,
    cli_runner,
    mocker,
):
    attach_persona("dummy@dummy.com", client_id="dummy-cluster")
    seed_clusters("dummy-cluster")
    test_app = make_test_app("list-all", list_all)
    mocked_render = mocker.patch("jobbergate_cli.subapps.clusters.app.terminal_message")
    result = cli_runner.invoke(test_app, ["list-all"])
    assert result.exit_code == 0, f"list-all failed: {result.stdout}"
    mocked_render.assert_called_once_with(
        "dummy-cluster",
        subject="Cluster Names",
        color="yellow",
        indent=True,
    )



import re
from unittest import mock

import pytest

from jobbergate_cli.auth import show_login_message, track_login_progress
from jobbergate_cli.context import JobbergateContext
from jobbergate_cli.config import settings
from jobbergate_cli.exceptions import Abort


@mock.patch("jobbergate_cli.context.Client")
def test_client_is_lazy_on_context(mocked_client, tweak_settings):
    ctx = JobbergateContext()

    mocked_client.assert_not_called()

    local_settings = dict(BASE_API_URL="test.example.com/", JOBBERGATE_REQUESTS_TIMEOUT=42)

    with tweak_settings(**local_settings):
        client = ctx.client

    assert client == mocked_client.return_value

    mocked_client.assert_called_once_with(
        base_url=local_settings["BASE_API_URL"],
        auth=ctx.authentication_handler,
        timeout=local_settings["JOBBERGATE_REQUESTS_TIMEOUT"],
    )


@mock.patch("jobbergate_cli.context.JobbergateAuthHandler")
def test_authentication_handler_is_lazy_on_context(mocked_auth_handler, tweak_settings):
    ctx = JobbergateContext()

    mocked_auth_handler.assert_not_called()

    local_settings = dict(
        OIDC_USE_HTTPS=True,
        OIDC_DOMAIN="example.com",
        OIDC_CLIENT_ID="client-id",
        OIDC_CLIENT_SECRET="client-secret",
    )

    with tweak_settings(**local_settings):
        auth_handler = ctx.authentication_handler

    assert auth_handler == mocked_auth_handler.return_value

    mocked_auth_handler.assert_called_once_with(
        cache_directory=settings.JOBBERGATE_USER_TOKEN_DIR,
        login_domain=f"https://{local_settings['OIDC_DOMAIN']}",
        login_client_id=local_settings["OIDC_CLIENT_ID"],
        login_client_secret=local_settings["OIDC_CLIENT_SECRET"],
        login_url_handler=show_login_message,
        login_sequence_handler=track_login_progress,
    )


@mock.patch("jobbergate_cli.context.JobbergateAuthHandler")
@pytest.mark.parametrize("missing_setting", ["OIDC_DOMAIN", "OIDC_CLIENT_ID"])
def test_authentication_handler__fails_on_missing_setting(mocked_auth_handler, missing_setting, tweak_settings):
    ctx = JobbergateContext()

    local_settings = dict(
        OIDC_DOMAIN="example.com",
        OIDC_CLIENT_ID="client-id",
        OIDC_CLIENT_SECRET="client-secret",
    )

    local_settings[missing_setting] = None

    with tweak_settings(**local_settings):
        with pytest.raises(
            Abort,
            match=re.compile(
                f"The following settings are required to enable authenticated requests:.*{missing_setting}$",
                flags=re.DOTALL,
            ),
        ):
            ctx.authentication_handler

    mocked_auth_handler.assert_not_called()



import shlex

from jobbergate_cli.main import main, login, logout


def test_main_command_with_ignore_username_and_password(make_test_app, cli_runner):
    test_app = make_test_app("jobbergate", main)

    result = cli_runner.invoke(test_app, shlex.split("jobbergate --username user --password pass"))
    assert result.exit_code == 0

    result = cli_runner.invoke(test_app, shlex.split("jobbergate -u user -p pass"))
    assert result.exit_code == 0


def test_login_success(make_test_app, cli_runner, dummy_context, attach_persona, mocker):
    test_app = make_test_app("login", login)

    mocked_login = mocker.patch.object(dummy_context.authentication_handler, "login")

    attach_persona("dummy@dummy.com")

    result = cli_runner.invoke(test_app, shlex.split("login"))
    assert result.exit_code == 0

    assert mocked_login.call_count == 1
    assert "User was logged in with email 'dummy@dummy.com'" in result.stdout


def test_logout_success(make_test_app, cli_runner, dummy_context, mocker):
    test_app = make_test_app("logout", logout)

    mocked_logout = mocker.patch.object(dummy_context.authentication_handler, "logout")

    result = cli_runner.invoke(test_app, shlex.split("logout"))
    assert result.exit_code == 0

    assert mocked_logout.call_count == 1
    assert "User was logged out" in result.stdout



import pytest

from jobbergate_cli.auth import (
    Console,
    DeviceCodeData,
    _show_login_narrow_screen,
    _show_login_standard_screen,
    open_on_browser,
    show_login_message,
    webbrowser,
)


@pytest.fixture(autouse=True)
def mocked_open_on_browser(mocker):
    """
    Provide a fixture that mocks the ``open_on_browser()`` function.
    """
    mocked = mocker.patch("jobbergate_cli.auth.open_on_browser")
    mocked.return_value = True
    return mocked


class TestOpenOnBrowser:
    def test_open_on_browser_valid_url(self, mocker):
        """
        Test that open_on_browser returns True when a valid browser is available.
        """
        mock_browser = mocker.Mock()
        mock_browser.open.return_value = True
        mocker.patch("webbrowser.get", return_value=mock_browser)

        result = open_on_browser("https://example.com")
        assert result is True
        mock_browser.open.assert_called_once_with("https://example.com")

    def test_open_on_browser_invalid_browser(self, mocker):
        """
        Test that open_on_browser returns False when a GenericBrowser is used.
        """
        mock_browser = webbrowser.GenericBrowser("dummy")
        mocker.patch("webbrowser.get", return_value=mock_browser)

        result = open_on_browser("https://example.com")
        assert result is False

    def test_open_on_browser_exception(self, mocker):
        """
        Test that open_on_browser returns False when an exception is raised.
        """
        mocker.patch("webbrowser.get", side_effect=Exception("Test exception"))

        result = open_on_browser("https://example.com")
        assert result is False


class TestShowLoginMessage:
    @pytest.fixture()
    def mocked_helpers(self, mocker):
        mocker.patch("jobbergate_cli.auth.open_on_browser", return_value=False)
        mocker.patch("jobbergate_cli.auth.copy_to_clipboard", return_value=False)

    def test_show_login_message__standard_screen(self, mocker, mocked_helpers):
        """
        Assert that the ``show_login_message()`` function will call ``_show_login_standard_screen()``.
        """
        verification_uri = "https://example.com"
        console_width = len(verification_uri) + 7

        device_code_data = DeviceCodeData(
            verification_uri_complete=verification_uri, expires_in=60, device_code="1234", interval=5
        )

        mocked_console = mocker.MagicMock()
        mocked_console.width = console_width
        mocker.patch("jobbergate_cli.auth.Console", return_value=mocked_console)

        mocked_show_on_narrow_screen = mocker.patch("jobbergate_cli.auth._show_login_narrow_screen")
        mocked_show_on_standard_screen = mocker.patch("jobbergate_cli.auth._show_login_standard_screen")

        show_login_message(device_code_data)

        assert mocked_show_on_narrow_screen.call_count == 0
        mocked_show_on_standard_screen.assert_called_once_with(verification_uri, 1)

        _show_login_standard_screen(verification_uri, 1)

    def test_show_login_message__narrow_screen(self, mocker, mocked_helpers):
        """
        Assert that the ``show_login_message()`` function will call ``_show_login_narrow_screen()``.
        """
        verification_uri = "https://example.com"
        console_width = len(verification_uri) + 7 - 1

        device_code_data = DeviceCodeData(
            verification_uri_complete=verification_uri, expires_in=60, device_code="1234", interval=5
        )

        mocked_console = mocker.MagicMock()
        mocked_console.width = console_width
        mocker.patch("jobbergate_cli.auth.Console", return_value=mocked_console)

        mocked_show_on_narrow_screen = mocker.patch("jobbergate_cli.auth._show_login_narrow_screen")
        mocked_show_on_standard_screen = mocker.patch("jobbergate_cli.auth._show_login_standard_screen")

        show_login_message(device_code_data)

        assert mocked_show_on_standard_screen.call_count == 0
        mocked_show_on_narrow_screen.assert_called_once_with(verification_uri, 1, mocked_console)

        _show_login_narrow_screen(verification_uri, 1, Console())



import importlib
import json
import pathlib
from textwrap import dedent
from unittest import mock

import httpx
import pytest
import respx

from jobbergate_cli.exceptions import Abort, JobbergateCliError
from jobbergate_cli.schemas import ApplicationResponse, JobScriptResponse, LocalApplication
from jobbergate_cli.subapps.job_scripts.tools import (
    JobbergateConfig,
    download_job_script_files,
    fetch_job_script_data,
    get_template_output_name_mapping,
    inject_sbatch_params,
    question_helper,
    remove_prefix_suffix,
    render_job_script,
    render_job_script_locally,
    render_template,
    upload_job_script_files,
    validate_parameter_file,
)


def test_validate_parameter_file__success(tmp_path):
    parameter_path = tmp_path / "dummy.json"
    dummy_data = dict(
        foo="one",
        bar=2,
        baz=False,
    )
    parameter_path.write_text(json.dumps(dummy_data))
    assert validate_parameter_file(parameter_path) == dummy_data


def test_validate_parameter_file__fails_if_file_does_not_exist():
    with pytest.raises(Abort, match="does not exist"):
        validate_parameter_file(pathlib.Path("some/fake/path"))


def test_validate_parameter_file__fails_if_file_is_not_valid_json(tmp_path):
    parameter_path = tmp_path / "dummy.json"
    parameter_path.write_text("clearly not json")
    with pytest.raises(Abort, match="is not valid JSON"):
        validate_parameter_file(parameter_path)


def test_fetch_job_script_data__success(
    respx_mock,
    dummy_context,
    dummy_job_script_data,
    dummy_domain,
):
    respx_mock.get(f"{dummy_domain}/jobbergate/job-scripts/1").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_job_script_data[0],
        ),
    )
    job_script = fetch_job_script_data(dummy_context, 1)
    assert job_script == JobScriptResponse.model_validate(dummy_job_script_data[0])


def test_inject_sbatch_params():
    sbatch_params = ["--comment=some_comment", "--nice=-1", "-N 10"]

    job_script_data_as_string = dedent(
        """
        #!/bin/bash

        #SBATCH --job-name=rats
        #SBATCH --partition=debug
        #SBATCH --output=sample-%j.out

        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    )

    expected_result = dedent(
        """
        #!/bin/bash

        #SBATCH --job-name=rats
        #SBATCH --partition=debug
        #SBATCH --output=sample-%j.out

        # Sbatch params injected at rendering time
        #SBATCH --comment=some_comment
        #SBATCH --nice=-1
        #SBATCH -N 10

        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    )

    actual_result = inject_sbatch_params(job_script_data_as_string, sbatch_params)
    assert actual_result == expected_result


def test_inject_sbatch_params__no_sbatch_flag():
    sbatch_params = ["--comment=some_comment", "--nice=-1", "-N 10"]

    job_script_data_as_string = dedent(
        """
        #!/bin/bash

        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    )

    expected_result = dedent(
        """
        #!/bin/bash

        # Sbatch params injected at rendering time
        #SBATCH --comment=some_comment
        #SBATCH --nice=-1
        #SBATCH -N 10

        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    )

    actual_result = inject_sbatch_params(job_script_data_as_string, sbatch_params)
    assert actual_result == expected_result


def test_inject_sbatch_params__empty_list():
    sbatch_params = []

    job_script_data_as_string = dedent(
        """
        #!/bin/bash

        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    )

    expected_result = dedent(
        """
        #!/bin/bash

        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    )

    actual_result = inject_sbatch_params(job_script_data_as_string, sbatch_params)
    assert actual_result == expected_result


def test_render_template__success(tmp_path):
    template_path = tmp_path / "dummy.j2"
    template_path.write_text("{{ foo }} {{ bar }} {{ baz }}")

    parameters = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )

    expected_output = "FOO BAR BAZ"

    assert render_template(template_path, parameters) == expected_output


def test_render_template__fail_when_unable_to_render(tmp_path):
    template_path = tmp_path / "dummy.j2"
    template_path.write_text("{{ foo + 42 }}")

    parameters = dict(not_a_key="bla")

    with pytest.raises(Abort, match="Unable to render"):
        render_template(template_path, parameters)


def test_render_template__fail_sandbox_violation(tmp_path):
    template_path = tmp_path / "dummy.j2"
    template_path.write_text("{{ foo.__str__() }}")

    parameters = dict(foo="bla")

    with pytest.raises(Abort, match="Security errors raised when rendering"):
        render_template(template_path, parameters)


def test_render_template__fails_if_template_does_not_exist(tmp_path):
    non_exist = tmp_path / "does/not/exist"
    with pytest.raises(Abort, match=f"Template file {non_exist} does not exist"):
        render_template(non_exist, {})


def test_render_template__fails_if_template_is_invalid(tmp_path):
    invalid_template = tmp_path / "invalid.j2"
    invalid_template.write_text("{{ foo")

    with pytest.raises(Abort, match=f"Unable to process jinja template filename={invalid_template}"):
        render_template(invalid_template, {})


def test_render_job_script_locally__success(
    dummy_context,
    dummy_application_dir,
    dummy_render_class,
    tmp_path,
    mocker,
):
    expected_template_data = dedent(
        """
        #!/bin/python3

        #SBATCH -J dummy_job
        #SBATCH -t 60
        print("I am a very, very dumb job script")
        print(f"foo='FOO'")
        print(f"bar='BAR'")
        print(f"baz='BAZ'")
        """
    )
    mocker.patch("jobbergate_cli.subapps.job_scripts.tools.render_template", return_value=expected_template_data)

    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )
    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )

    write_mock = mocker.patch("builtins.open", mocker.mock_open())

    render_job_script_locally(dummy_context, "dummy-job-script", dummy_application_dir, tmp_path, fast=True)

    write_mock().write.assert_called_once_with(expected_template_data)


def test_render_job_script_locally__fail_when_no_workflow_file_specified(tmp_path, dummy_context, mocker):
    app_data_without_workflow = LocalApplication(
        template_vars={},
        template_files=[],
        workflow_files=[],
    )

    mocker.patch(
        "jobbergate_cli.subapps.job_scripts.tools.fetch_application_data_locally",
        return_value=app_data_without_workflow,
    )

    with pytest.raises(Abort, match="Application does not have a workflow file"):
        render_job_script_locally(dummy_context, "dummy-job-script", tmp_path, tmp_path, fast=True)


def test_render_job_script_locally__with_sbatch_params(
    dummy_render_class, dummy_context, dummy_application_dir, mocker
):
    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )
    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )

    render_job_script_locally(
        dummy_context,
        "dummy-job-script",
        dummy_application_dir,
        dummy_application_dir,
        sbatch_params=["--comment=some_comment", "--nice=-1", "-N 10"],
    )

    assert (dummy_application_dir / "job-script-template.py").read_text() == dedent(
        """\
        #!/bin/python3

        #SBATCH -J dummy_job
        #SBATCH -t 60
        # Sbatch params injected at rendering time
        #SBATCH --comment=some_comment
        #SBATCH --nice=-1
        #SBATCH -N 10

        print("I am a very, very dumb job script")
        print(f"foo='FOO'")
        print(f"bar='BAR'")
        print(f"baz='BAZ'")"""
    )


def test_render_job_script__providing_a_name(
    dummy_application_data,
    dummy_job_script_data,
    dummy_module_source,
    dummy_domain,
    dummy_context,
    dummy_render_class,
    attach_persona,
    respx_mock,
    mocker,
):
    """
    Test that we can create a job script with the desired name.
    """
    attach_persona("dummy@dummy.com")

    application_response = ApplicationResponse(**dummy_application_data[0])
    mocked_fetch_application_data = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.tools.fetch_application_data",
        return_value=application_response,
    )
    assert application_response.workflow_files is not None
    get_workflow_route = respx_mock.get(f"{dummy_domain}{application_response.workflow_files[0].path}")
    get_workflow_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            content=dummy_module_source.encode(),
        ),
    )

    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )

    desired_job_script_data = dummy_job_script_data[0]

    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )
    create_route = respx_mock.post(
        f"{dummy_domain}/jobbergate/job-scripts/render-from-template/{application_response.application_id}"
    )
    create_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=desired_job_script_data,
        ),
    )

    actual_job_script_data = render_job_script(
        dummy_context,
        id_or_identifier=1,
        name=desired_job_script_data["name"],
        fast=True,
    )

    mocked_fetch_application_data.assert_called_once_with(dummy_context, 1)

    assert actual_job_script_data == JobScriptResponse.model_validate(desired_job_script_data)


def test_render_job_script__set_name_dynamically_from_application_config(
    dummy_application_data,
    dummy_job_script_data,
    dummy_module_source,
    dummy_domain,
    dummy_context,
    dummy_render_class,
    attach_persona,
    respx_mock,
    mocker,
):
    """
    Test that we can create a job script by setting its name at runtime.
    """
    attach_persona("dummy@dummy.com")

    new_name = "new-name"

    application_response = ApplicationResponse(**dummy_application_data[0])
    application_response.template_vars["job_script_name"] = new_name

    mocked_fetch_application_data = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.tools.fetch_application_data",
        return_value=application_response,
    )
    assert application_response.workflow_files is not None
    get_workflow_route = respx_mock.get(f"{dummy_domain}{application_response.workflow_files[0].path}")
    get_workflow_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            content=dummy_module_source.encode(),
        ),
    )

    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )

    desired_job_script_data = dummy_job_script_data[0]

    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )
    create_route = respx_mock.post(
        f"{dummy_domain}/jobbergate/job-scripts/render-from-template/{application_response.application_id}"
    )
    create_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=desired_job_script_data,
        ),
    )

    actual_job_script_data = render_job_script(
        dummy_context,
        id_or_identifier=1,
        name=new_name,
        fast=True,
    )

    mocked_fetch_application_data.assert_called_once_with(dummy_context, 1)

    assert actual_job_script_data == JobScriptResponse.model_validate(desired_job_script_data)


def test_render_job_script__set_name_dynamically_from_jobbergate_config(
    dummy_application_data,
    dummy_job_script_data,
    dummy_module_source,
    dummy_domain,
    dummy_context,
    dummy_render_class,
    attach_persona,
    respx_mock,
    mocker,
):
    """
    Test that we can create a job script by setting its name at runtime.
    """
    attach_persona("dummy@dummy.com")

    new_name = "new-name"

    application_response = ApplicationResponse(**dummy_application_data[0])
    application_response.workflow_files[0].runtime_config["job_script_name"] = new_name

    mocked_fetch_application_data = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.tools.fetch_application_data",
        return_value=application_response,
    )
    assert application_response.workflow_files is not None
    get_workflow_route = respx_mock.get(f"{dummy_domain}{application_response.workflow_files[0].path}")
    get_workflow_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            content=dummy_module_source.encode(),
        ),
    )

    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )

    desired_job_script_data = dummy_job_script_data[0]

    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )
    create_route = respx_mock.post(
        f"{dummy_domain}/jobbergate/job-scripts/render-from-template/{application_response.application_id}"
    )
    create_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=desired_job_script_data,
        ),
    )

    actual_job_script_data = render_job_script(
        dummy_context,
        id_or_identifier=1,
        name=new_name,
        fast=True,
    )

    mocked_fetch_application_data.assert_called_once_with(dummy_context, 1)

    assert actual_job_script_data == JobScriptResponse.model_validate(desired_job_script_data)


def test_render_job_script__without_a_name(
    dummy_application_data,
    dummy_job_script_data,
    dummy_module_source,
    dummy_domain,
    dummy_context,
    dummy_render_class,
    attach_persona,
    respx_mock,
    mocker,
):
    """
    Test that we can create a job script without providing a name.

    In this case, it should be created with the name of the base application.
    """
    attach_persona("dummy@dummy.com")

    application_response = ApplicationResponse(**dummy_application_data[0])
    mocked_fetch_application_data = mocker.patch(
        "jobbergate_cli.subapps.job_scripts.tools.fetch_application_data",
        return_value=application_response,
    )
    assert application_response.workflow_files is not None
    get_workflow_route = respx_mock.get(f"{dummy_domain}{application_response.workflow_files[0].path}")
    get_workflow_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            content=dummy_module_source.encode(),
        ),
    )

    dummy_render_class.prepared_input = dict(
        foo="FOO",
        bar="BAR",
        baz="BAZ",
    )

    desired_job_script_data = dummy_job_script_data[0]
    desired_job_script_data["name"] = application_response.name

    mocker.patch.object(
        importlib.import_module("inquirer.prompt"),
        "ConsoleRender",
        new=dummy_render_class,
    )
    create_route = respx_mock.post(
        f"{dummy_domain}/jobbergate/job-scripts/render-from-template/{application_response.application_id}"
    )
    create_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=desired_job_script_data,
        ),
    )

    actual_job_script_data = render_job_script(
        dummy_context,
        id_or_identifier=application_response.application_id,
        name=None,
        fast=True,
    )

    mocked_fetch_application_data.assert_called_once_with(dummy_context, application_response.application_id)

    assert actual_job_script_data == JobScriptResponse.model_validate(desired_job_script_data)


def test_question_helper__return_actual_value_when_actual_value_is_not_none():
    question_func = mock.Mock()

    assert (
        question_helper(
            question_func=question_func,
            text="Give me foo",
            default="foo",
            fast=True,
            actual_value="bar",
        )
        == "bar"
    )

    assert question_func.call_count == 0

    assert (
        question_helper(
            question_func=question_func,
            text="Give me foo",
            default="foo",
            fast=False,
            actual_value="bar",
        )
        == "bar"
    )

    assert question_func.call_count == 0


def test_question_helper__return_default_when_actual_value_is_none_on_fast_mode():
    question_func = mock.Mock()

    assert (
        question_helper(
            question_func=question_func,
            text="Give me foo",
            default="foo",
            fast=True,
            actual_value=None,
        )
        == "foo"
    )

    assert question_func.call_count == 0


def test_question_helper__ask_question_when_actual_value_is_none_on_non_fast_mode():
    question_func = mock.Mock()

    question_helper(
        question_func=question_func,
        text="Give me foo",
        default="foo",
        fast=False,
        actual_value=None,
    )

    question_func.assert_called_once_with("Give me foo", default="foo")


class TestDownloadJobScriptFiles:
    """
    Test the save_job_script_files function.
    """

    def test_download_job_scripts_files__all_files(
        self,
        tmp_path,
        respx_mock,
        dummy_context,
        dummy_domain,
        dummy_job_script_data,
        dummy_template_source,
    ):
        """
        Test that we can download all the files from a job script.
        """
        job_script_id = 1
        respx_mock.get(f"{dummy_domain}/jobbergate/job-scripts/{job_script_id}").mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                json=dummy_job_script_data[0],
            ),
        )
        job_script_data = JobScriptResponse.model_validate(dummy_job_script_data[0])

        get_file_routes = [respx_mock.get(f"{dummy_domain}{f.path}") for f in job_script_data.files]
        for route in get_file_routes:
            route.mock(
                return_value=httpx.Response(
                    httpx.codes.OK,
                    content=dummy_template_source.encode(),
                ),
            )

        list_of_files = [tmp_path / f.filename for f in job_script_data.files]

        assert len(list_of_files) >= 1, "sanity check"

        files = download_job_script_files(job_script_id, dummy_context, tmp_path)

        assert files == job_script_data.files
        assert set(tmp_path.rglob("*")) == set(list_of_files)
        assert all(r.called for r in get_file_routes)

        assert all(p.read_text() == dummy_template_source for p in list_of_files)


@pytest.mark.parametrize(
    "input_string, expected_output",
    [
        ("templates/test.sh.j2", "test.sh"),
        ("path/to/file.py.j2", "path/to/file.py"),
        ("templates/file.py.jinja2", "file.py"),
        ("templates/test.sh.jinja2", "test.sh"),
        ("other/path/test.py", "other/path/test.py"),
    ],
)
def test_remove_prefix_suffix(input_string, expected_output):
    assert remove_prefix_suffix(input_string) == expected_output


class TestGetTemplateOutputNameMapping:
    """Test the get_template_output_name_mapping function."""

    def test_default_template_valid_output_name(self):
        config = JobbergateConfig(
            default_template="templates/template.j2",
            supporting_files=None,
            supporting_files_output_name=None,
        )
        expected_mapping = {"template.j2": "template"}

        assert get_template_output_name_mapping(config, "dummy-name") == expected_mapping

    def test_default_template_valid_output_name__legacy_mode(self, tweak_settings):
        config = JobbergateConfig(
            default_template="templates/template.j2",
            supporting_files=None,
            supporting_files_output_name=None,
        )
        expected_mapping = {"template.j2": "dummy-name.job"}

        with tweak_settings(JOBBERGATE_LEGACY_NAME_CONVENTION=True):
            assert get_template_output_name_mapping(config, "dummy-name") == expected_mapping

    def test_supporting_files_with_valid_output_names(self):
        config = JobbergateConfig(
            template_files=[pathlib.Path("templates/template1.j2"), pathlib.Path("templates/template2.j2")],
            default_template="templates/template1.j2",
            supporting_files=["templates/support1.j2", "templates/support2.j2"],
            supporting_files_output_name={
                "templates/support1.j2": ["output1.txt"],
                "templates/support2.j2": ["output2.txt"],
            },
        )

        expected_mapping = {"template1.j2": "template1", "support1.j2": "output1.txt", "support2.j2": "output2.txt"}

        assert get_template_output_name_mapping(config, "dummy-name") == expected_mapping

    def test_default_template_not_specified(self):
        config = JobbergateConfig(
            template_files=[],
            supporting_files_output_name=None,
            supporting_files=None,
        )
        with pytest.raises(Abort, match="Default template was not specified"):
            get_template_output_name_mapping(config, "dummy-name")

    def test_supporting_files_output_names_multiple_values(self):
        config = JobbergateConfig(
            default_template="templates/template1.j2",
            supporting_files_output_name={
                "template2.j2": ["output3", "output4"],
            },
        )

        with pytest.raises(Abort, match="template='template2.j2' has 2 output names"):
            get_template_output_name_mapping(config, "dummy-name")


class TestUploadJobScriptFiles:
    """Test the upload_job_script_files function."""

    job_script_id = 1

    @pytest.fixture(scope="function")
    def mocked_routes(self, dummy_domain):
        def _helper(assert_all_mocked=True, assert_all_called=False):
            app_mock = respx.mock(
                base_url=dummy_domain,
                assert_all_mocked=assert_all_mocked,
                assert_all_called=assert_all_called,
            )

            app_mock.put(
                path=f"/jobbergate/job-scripts/{self.job_script_id}/upload/ENTRYPOINT",
                name="upload_entrypoint",
            ).respond(httpx.codes.OK)

            app_mock.put(
                path=f"/jobbergate/job-scripts/{self.job_script_id}/upload/SUPPORT",
                name="upload_support",
            ).respond(httpx.codes.OK)
            return app_mock

        return _helper

    def test_upload_job_script__success(
        self,
        dummy_context,
        mocked_routes,
        tmp_path,
    ):
        with mocked_routes(assert_all_called=True) as routes:
            dummy_job_script = tmp_path / "dummy.sh"
            dummy_job_script.write_text("echo hello world")

            dummy_support_1 = tmp_path / "dummy-support-1.txt"
            dummy_support_1.write_text("dummy 1")

            dummy_support_2 = tmp_path / "dummy-support-2.txt"
            dummy_support_2.write_text("dummy 2")

            upload_job_script_files(
                dummy_context, self.job_script_id, dummy_job_script, [dummy_support_1, dummy_support_2]
            )

            assert routes["upload_entrypoint"].call_count == 1
            assert b'filename="dummy.sh"' in routes["upload_entrypoint"].calls[0].request.content
            assert b"echo hello world" in routes["upload_entrypoint"].calls[0].request.content

            assert routes["upload_support"].call_count == 2

            assert b'filename="dummy-support-1.txt"' in routes["upload_support"].calls[0].request.content
            assert b"dummy 1" in routes["upload_support"].calls[0].request.content

            assert b'filename="dummy-support-2.txt"' in routes["upload_support"].calls[1].request.content
            assert b"dummy 2" in routes["upload_support"].calls[1].request.content

    def test_upload_job_script__raises_exception_if_context_client_is_undefined(
        self,
        dummy_context,
        tmp_path,
    ):
        dummy_job_script = tmp_path / "dummy.sh"
        dummy_job_script.write_text("echo hello world")

        dummy_context.client = None
        with pytest.raises(JobbergateCliError, match="not defined"):
            upload_job_script_files(dummy_context, self.job_script_id, dummy_job_script)

    def test_upload_job_script__raises_exception_if_job_script_does_not_exist_or_is_not_a_file(
        self,
        dummy_context,
        tmp_path,
    ):
        dummy_job_script = tmp_path / "dummy.sh"

        with pytest.raises(Abort, match="Job Script file .* does not exist"):
            upload_job_script_files(dummy_context, self.job_script_id, dummy_job_script)

        dummy_job_script.mkdir()
        with pytest.raises(Abort, match="Job Script file .* is not a file"):
            upload_job_script_files(dummy_context, self.job_script_id, dummy_job_script)

    def test_upload_job_script__raises_exception_if_job_supporting_file_does_not_exist_or_is_not_a_file(
        self,
        dummy_context,
        tmp_path,
    ):
        dummy_job_script = tmp_path / "dummy.sh"
        dummy_job_script.touch()

        dummy_support_1 = tmp_path / "dummy-support-1.txt"

        with pytest.raises(Abort, match="Supporting file .* does not exist"):
            upload_job_script_files(dummy_context, self.job_script_id, dummy_job_script, [dummy_support_1])

        dummy_support_1.mkdir()
        with pytest.raises(Abort, match="Supporting file .* is not a file"):
            upload_job_script_files(dummy_context, self.job_script_id, dummy_job_script, [dummy_support_1])

    def test_upload_job_script__raises_exception_if_job_script_fails_to_upload(
        self,
        dummy_context,
        mocked_routes,
        tmp_path,
    ):
        with mocked_routes() as routes:
            dummy_job_script = tmp_path / "dummy.sh"
            dummy_job_script.write_text("echo hello world")
            routes["upload_entrypoint"].respond(httpx.codes.BAD_REQUEST)

            with pytest.raises(Abort, match="Job Script file .* failed to upload"):
                upload_job_script_files(dummy_context, self.job_script_id, dummy_job_script)

            assert routes["upload_entrypoint"].call_count == 1
            assert b'filename="dummy.sh"' in routes["upload_entrypoint"].calls[0].request.content

    def test_upload_job_script__raises_exception_if_any_of_the_supporting_files_fail_to_upload(
        self,
        dummy_context,
        mocked_routes,
        tmp_path,
    ):
        with mocked_routes(assert_all_called=True) as routes:
            routes["upload_support"].respond(httpx.codes.BAD_REQUEST)

            dummy_job_script = tmp_path / "dummy.sh"
            dummy_job_script.write_text("echo hello world")

            dummy_support_1 = tmp_path / "dummy-support-1.txt"
            dummy_support_1.write_text("dummy 1")

            dummy_support_2 = tmp_path / "dummy-support-2.txt"
            dummy_support_2.write_text("dummy 2")

            with pytest.raises(JobbergateCliError, match="Supporting file .* was not accepted"):
                upload_job_script_files(
                    dummy_context, self.job_script_id, dummy_job_script, [dummy_support_1, dummy_support_2]
                )

            assert routes["upload_entrypoint"].call_count == 1
            assert b'filename="dummy.sh"' in routes["upload_entrypoint"].calls[0].request.content

            assert routes["upload_support"].call_count == 2
            assert b'filename="dummy-support-1.txt"' in routes["upload_support"].calls[0].request.content
            assert b'filename="dummy-support-2.txt"' in routes["upload_support"].calls[1].request.content



import pathlib
from unittest import mock

import httpx
import pytest

from jobbergate_cli.exceptions import Abort
from jobbergate_cli.schemas import JobScriptResponse, JobSubmissionCreateRequestData, JobSubmissionResponse
from jobbergate_cli.subapps.job_submissions.tools import (
    OnsiteJobSubmission,
    RemoteJobSubmission,
    fetch_job_submission_data,
    job_submissions_factory,
)


@pytest.mark.parametrize("submission_cls", [OnsiteJobSubmission, RemoteJobSubmission])
@pytest.mark.parametrize("organization_id", [None, "", "some-organization"])
class TestJobSubmissionsClusterName:
    def test_with_explicit_cluster_name(
        self, dummy_context, attach_persona, submission_cls, organization_id, tweak_settings
    ):
        cluster_name = "test-cluster"
        attach_persona("dummy@dummy.com", organization_id=organization_id)
        with tweak_settings(DEFAULT_CLUSTER_NAME="default-cluster"):
            submission_handler = submission_cls(
                jg_ctx=dummy_context,
                job_script_id=1,
                name="test",
                cluster_name=cluster_name,
            )
        expected_cluster_name = cluster_name
        if organization_id:
            expected_cluster_name += f"-{organization_id}"
        assert submission_handler.cluster_name == expected_cluster_name

    def test_default_cluster_name(self, dummy_context, attach_persona, submission_cls, organization_id, tweak_settings):
        cluster_name = "default-cluster"
        attach_persona("dummy@dummy.com", organization_id=organization_id)
        with tweak_settings(DEFAULT_CLUSTER_NAME=cluster_name):
            submission_handler = submission_cls(
                jg_ctx=dummy_context,
                job_script_id=1,
                name="test",
            )
        expected_cluster_name = cluster_name
        if organization_id:
            expected_cluster_name += f"-{organization_id}"
        assert submission_handler.cluster_name == expected_cluster_name

    def test_throws_exception_with_no_explicit_or_default_cluster_name(
        self, dummy_context, attach_persona, submission_cls, organization_id, tweak_settings
    ):
        attach_persona("dummy@dummy.com", organization_id=organization_id)
        with (
            tweak_settings(DEFAULT_CLUSTER_NAME=None),
            pytest.raises(ValueError, match="No cluster name supplied and no default exists"),
        ):
            submission_cls(
                jg_ctx=dummy_context,
                job_script_id=1,
                name="test",
            )

    def test_with_organization_id_on_cluster_name(
        self, dummy_context, attach_persona, submission_cls, organization_id, tweak_settings
    ):
        cluster_name = "test-cluster"
        if organization_id:
            cluster_name += f"-{organization_id}"
        attach_persona("dummy@dummy.com", organization_id=organization_id)
        with tweak_settings(DEFAULT_CLUSTER_NAME="default-cluster"):
            submission_handler = submission_cls(
                jg_ctx=dummy_context,
                job_script_id=1,
                name="test",
                cluster_name=cluster_name,
            )
        assert submission_handler.cluster_name == cluster_name


@pytest.mark.parametrize("submission_cls", [OnsiteJobSubmission, RemoteJobSubmission])
@pytest.mark.parametrize(
    "execution_directory", [pathlib.Path("./some/relative/path"), pathlib.Path("/some/absolute/path"), None]
)
class TestJobSubmissionsExecutionDirectory:
    def test_execution_directory_is_absolute(self, dummy_context, attach_persona, submission_cls, execution_directory):
        attach_persona("dummy@dummy.com")

        submission_handler = submission_cls(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            execution_directory=execution_directory,
            cluster_name="test-cluster",
        )

        assert submission_handler.execution_directory.is_absolute()

        if execution_directory is None:
            # Default value
            execution_directory = pathlib.Path.cwd()

        expected_execution_directory = execution_directory.resolve()
        assert submission_handler.execution_directory == expected_execution_directory


class TestJobSubmissionsGetRequestData:
    def test_handle_request_data__on_site(self, dummy_context, attach_persona):
        attach_persona("dummy@dummy.com")

        dummy_data = dict(
            job_script_id=1,
            name="test",
            execution_directory=pathlib.Path("/some/fake/path"),
            sbatch_arguments=["--partition=debug", "--time=1:00:00"],
            description="test description",
            cluster_name="test-cluster",
        )

        submission_handler = OnsiteJobSubmission(jg_ctx=dummy_context, download=True, **dummy_data)

        # simulate process_submission by setting slurm_job_id
        dummy_data["slurm_job_id"] = 1234
        submission_handler.slurm_job_id = dummy_data["slurm_job_id"]

        actual_request_data = submission_handler.get_request_data()
        expected_request_data = JobSubmissionCreateRequestData.model_validate(dummy_data)

        assert actual_request_data == expected_request_data

    def test_handle_request_data__remote(self, dummy_context, attach_persona, dummy_domain, respx_mock):
        attach_persona("dummy@dummy.com")

        dummy_data = dict(
            job_script_id=1,
            name="test",
            execution_directory=pathlib.Path("/some/fake/path"),
            sbatch_arguments=["--partition=debug", "--time=1:00:00"],
            description="test description",
            cluster_name="test-cluster",
        )

        submission_handler = RemoteJobSubmission(jg_ctx=dummy_context, download=True, **dummy_data)

        actual_request_data = submission_handler.get_request_data()
        expected_request_data = JobSubmissionCreateRequestData.model_validate(dummy_data)

        assert actual_request_data == expected_request_data


@pytest.mark.parametrize("submission_cls", [OnsiteJobSubmission, RemoteJobSubmission])
class TestJobSubmissionsMakePostRequest:
    def test_post_request__success(
        self, dummy_context, attach_persona, submission_cls, dummy_job_submission_data, respx_mock, dummy_domain
    ):
        attach_persona("dummy@dummy.com")
        job_submission_data = dummy_job_submission_data[0]

        submission_handler = submission_cls(
            jg_ctx=dummy_context,
            job_script_id=job_submission_data["job_script_id"],
            name=job_submission_data["name"],
            cluster_name="test-cluster",
        )

        create_job_submission_data = JobSubmissionCreateRequestData.model_validate(job_submission_data)
        create_job_submission_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-submissions")
        create_job_submission_route.mock(
            return_value=httpx.Response(
                httpx.codes.CREATED,
                json=job_submission_data,
            ),
        )

        actual_response = submission_handler.make_post_request(create_job_submission_data)
        expected_response = JobSubmissionResponse.model_validate(job_submission_data)

        assert actual_response == expected_response


class TestJobSubmissionsProcessSubmissions:
    def test_process_submission__remote_download(self, dummy_context, mocker, attach_persona):
        attach_persona("dummy@dummy.com")
        submission_handler = RemoteJobSubmission(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
            download=True,
        )
        mocked_download_job_script_files = mocker.patch(
            "jobbergate_cli.subapps.job_submissions.tools.download_job_script_files"
        )
        submission_handler.process_submission()
        mocked_download_job_script_files.assert_called_once_with(
            submission_handler.job_script_id, submission_handler.jg_ctx, submission_handler.execution_directory
        )

    def test_process_submission__no_remote_download(self, dummy_context, mocker, attach_persona):
        attach_persona("dummy@dummy.com")
        submission_handler = RemoteJobSubmission(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
            download=False,
        )
        mocked_download_job_script_files = mocker.patch(
            "jobbergate_cli.subapps.job_submissions.tools.download_job_script_files"
        )
        submission_handler.process_submission()
        mocked_download_job_script_files.assert_not_called()

    def test_process_submission__on_site_abort_if_sbatch_path_is_unset(
        self, dummy_context, attach_persona, tweak_settings
    ):
        attach_persona("dummy@dummy.com")
        submission_handler = OnsiteJobSubmission(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
        )

        with (
            tweak_settings(SBATCH_PATH=None),
            pytest.raises(Abort, match="SBATCH_PATH most be set for onsite submissions"),
        ):
            submission_handler.process_submission()

    class DummyFile:
        @property
        def file_type(self):
            return "NOT-ENTRYPOINT"

    @pytest.mark.parametrize("job_script_files", [[], [DummyFile(), DummyFile()]])
    def test_process_submission__on_site_abort_if_not_exact_one_entrypoint_file_is_found(
        self, job_script_files, dummy_context, attach_persona, tweak_settings, mocker, tmp_path
    ):
        attach_persona("dummy@dummy.com")
        submission_handler = OnsiteJobSubmission(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
        )

        mocker.patch(
            "jobbergate_cli.subapps.job_submissions.tools.download_job_script_files", return_value=job_script_files
        )

        with (
            tweak_settings(SBATCH_PATH=tmp_path),
            pytest.raises(Abort, match="There should be exactly one entrypoint file in the parent job script"),
        ):
            submission_handler.process_submission()

    @pytest.mark.parametrize("download", [True, False])
    def test_process_submission__on_site_success(
        self, download, dummy_context, attach_persona, tweak_settings, mocker, tmp_path, dummy_job_script_data
    ):
        attach_persona("dummy@dummy.com")

        job_script_data = JobScriptResponse.model_validate(dummy_job_script_data[0])

        submission_handler = OnsiteJobSubmission(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
            download=download,
            execution_directory=tmp_path,
        )

        mocked_download_job_script_files = mocker.patch(
            "jobbergate_cli.subapps.job_submissions.tools.download_job_script_files", return_value=job_script_data.files
        )
        mocked_inject_sbatch_params = mocker.patch.object(submission_handler, "inject_sbatch_params")

        mocked_sbatch = mock.MagicMock()
        mocked_sbatch.submit_job = lambda *args, **kwargs: 13
        mocker.patch("jobbergate_cli.subapps.job_submissions.tools.SubmissionHandler", return_value=mocked_sbatch)

        with tweak_settings(SBATCH_PATH=tmp_path):
            submission_handler.process_submission()

        assert submission_handler.slurm_job_id == 13

        # files are downloaded anyway for on-site submissions
        mocked_download_job_script_files.assert_called_once_with(1, dummy_context, tmp_path)
        assert mocked_inject_sbatch_params.call_count == 1

    def test_inject_sbatch_params__on_site(self, mocker, attach_persona, dummy_context, tmp_path):
        attach_persona("dummy@dummy.com")
        submission_handler = OnsiteJobSubmission(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
            execution_directory=tmp_path,
            sbatch_arguments=["--partition=debug", "--time=1:00:00"],
        )

        job_script_path = tmp_path / "entrypoint.sh"

        job_script_path.write_text("original content")

        mocked_inject_sbatch_params = mocker.patch(
            "jobbergate_cli.subapps.job_submissions.tools.inject_sbatch_params", return_value="inject_sbatch_params"
        )

        submission_handler.inject_sbatch_params(job_script_path)

        mocked_inject_sbatch_params.assert_called_once_with(
            "original content", ["--partition=debug", "--time=1:00:00"], "Injected at submission time by Jobbergate CLI"
        )
        assert job_script_path.read_text() == "inject_sbatch_params"

    def test_skip_inject_sbatch_params__on_site(self, mocker, attach_persona, dummy_context, tmp_path):
        attach_persona("dummy@dummy.com")
        submission_handler = OnsiteJobSubmission(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
            execution_directory=tmp_path,
            sbatch_arguments=[],
        )

        job_script_path = tmp_path / "entrypoint.sh"

        job_script_path.write_text("original content")

        mocked_inject_sbatch_params = mocker.patch("jobbergate_cli.subapps.job_submissions.tools.inject_sbatch_params")

        submission_handler.inject_sbatch_params(job_script_path)

        assert mocked_inject_sbatch_params.call_count == 0
        assert job_script_path.read_text() == "original content"


@pytest.mark.parametrize("submission_cls", [OnsiteJobSubmission, RemoteJobSubmission])
class TestJobSubmissionsRun:
    def test_run__success(self, dummy_context, attach_persona, submission_cls, mocker):
        attach_persona("dummy@dummy.com")
        submission_handler = submission_cls(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
        )

        mocked_process_submission = mocker.patch.object(submission_handler, "process_submission")
        mocked_get_request_data = mocker.patch.object(
            submission_handler, "get_request_data", return_value="request_data"
        )
        mocked_make_post_request = mocker.patch.object(submission_handler, "make_post_request", return_value="response")

        actual_response = submission_handler.run()

        assert actual_response == "response"
        mocked_process_submission.assert_called_once()
        mocked_get_request_data.assert_called_once()
        mocked_make_post_request.assert_called_once_with("request_data")


@pytest.mark.parametrize(
    "submission_cls,sbatch_path", [[OnsiteJobSubmission, "/usr/bin/sbatch"], [RemoteJobSubmission, None]]
)
def test_job_submissions_factory(submission_cls, sbatch_path, attach_persona, dummy_context, tweak_settings):
    attach_persona("dummy@dummy.com")
    with tweak_settings(SBATCH_PATH=sbatch_path):
        submission_handler = job_submissions_factory(
            jg_ctx=dummy_context,
            job_script_id=1,
            name="test",
            cluster_name="test-cluster",
            download=True,
        )
    assert isinstance(submission_handler, submission_cls)
    assert submission_handler.jg_ctx == dummy_context
    assert submission_handler.job_script_id == 1
    assert submission_handler.name == "test"
    assert submission_handler.cluster_name == "test-cluster"
    assert submission_handler.download is True


def test_fetch_job_submission_data__success__using_id(
    respx_mock,
    dummy_context,
    dummy_job_submission_data,
    dummy_domain,
):
    job_submission_data = dummy_job_submission_data[3]
    job_submission_id = job_submission_data["id"]
    fetch_route = respx_mock.get(f"{dummy_domain}/jobbergate/job-submissions/{job_submission_id}")
    fetch_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=job_submission_data,
        ),
    )

    result = fetch_job_submission_data(dummy_context, job_submission_id)
    assert fetch_route.called
    assert result == JobSubmissionResponse(**job_submission_data)
    assert result.report_message == job_submission_data.get("report_message")



import shlex
from unittest import mock

import httpx
import pytest

from jobbergate_cli.exceptions import Abort
from jobbergate_cli.schemas import ApplicationResponse
from jobbergate_cli.subapps.applications.app import (
    HIDDEN_FIELDS,
    clone,
    create,
    delete,
    download_files,
    get_one,
    list_all,
    pathlib,
    style_mapper,
    update,
)
from jobbergate_cli.text_tools import unwrap


def test_list_all__renders_paginated_results(
    make_test_app,
    dummy_context,
    cli_runner,
    mocker,
):
    test_app = make_test_app("list-all", list_all)
    mocked_pagination = mocker.patch("jobbergate_cli.subapps.applications.app.handle_pagination")
    result = cli_runner.invoke(test_app, ["list-all"])
    assert result.exit_code == 0, f"list-all failed: {result.output}"
    mocked_pagination.assert_called_once_with(
        jg_ctx=dummy_context,
        url_path="/jobbergate/job-script-templates",
        abort_message="Couldn't retrieve applications list from API",
        params={
            "include_null_identifier": False,
            "user_only": False,
            "sort_ascending": False,
            "sort_field": "id",
            "include_archived": False,
        },
        title="Applications List",
        style_mapper=style_mapper,
        hidden_fields=HIDDEN_FIELDS,
        nested_response_model_cls=ApplicationResponse,
    )


@pytest.mark.parametrize(
    "selector_template",
    [
        "{id}",
        "-i {id}",
        "--id={id}",
        "--id {id}",
        "{identifier}",
        "--identifier={identifier}",
        "--identifier {identifier}",
    ],
)
def test_get_one__success(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_application_data,
    dummy_domain,
    cli_runner,
    mocker,
    selector_template,
):
    application_data = dummy_application_data[0]
    id = application_data["id"]
    identifier = application_data["identifier"]

    url_selector = identifier if "identifier" in selector_template else id
    cli_selector = selector_template.format(id=id, identifier=identifier)

    respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates/{url_selector}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dummy_application_data[0],
        ),
    )
    test_app = make_test_app("get-one", get_one)
    mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.render_single_result")
    result = cli_runner.invoke(test_app, shlex.split(f"get-one {cli_selector}"))
    assert result.exit_code == 0, f"get-one failed: {result.output}"
    mocked_render.assert_called_once_with(
        dummy_context,
        ApplicationResponse(**dummy_application_data[0]),
        title="Application",
        hidden_fields=HIDDEN_FIELDS,
    )


def test_get_one__fails_with_neither_id_or_identifier(make_test_app, cli_runner):
    test_app = make_test_app("get-one", get_one)
    result = cli_runner.invoke(test_app, shlex.split("get-one"))
    assert result.exit_code != 0
    assert "You must supply one and only one selection value" in result.output


@pytest.mark.parametrize(
    "cli_selector",
    ["1 --id 2", "foo --identifier bar", "--id 1 --identifier dummy"],
)
def test_get_one__fails_with_both_id_and_identifier(make_test_app, cli_runner, cli_selector):
    test_app = make_test_app("get-one", get_one)
    result = cli_runner.invoke(test_app, shlex.split(f"get-one {cli_selector}"))
    assert result.exit_code != 0
    assert "You must supply one and only one selection value" in result.output


@pytest.mark.parametrize(
    "name_flag,application_path_flag,separator", [("--name", "--application-path", "="), ("-n", "-a", " ")]
)
def test_create__success(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_application_data,
    dummy_application_dir,
    dummy_domain,
    cli_runner,
    mocker,
    name_flag,
    application_path_flag,
    separator,
):
    response_data = dummy_application_data[0]
    response_data["application_uploaded"] = False

    create_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-script-templates")
    create_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=response_data,
        ),
    )

    mocked_upload = mocker.patch("jobbergate_cli.subapps.applications.app.upload_application")

    test_app = make_test_app("create", create)
    mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.render_single_result")
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                create {name_flag}{separator}dummy-name --identifier=dummy-identifier
                       {application_path_flag}{separator}{dummy_application_dir}
                       --application-desc="This application is kinda dumb, actually"
                """
            )
        ),
    )
    assert result.exit_code == 0, f"create failed: {result.output}"
    assert create_route.called
    assert mocked_upload.called

    mocked_render.assert_called_once_with(
        dummy_context,
        {**response_data, "application_uploaded": True},
        title="Created Application",
        hidden_fields=HIDDEN_FIELDS,
    )


def test_create__warns_but_does_not_abort_if_upload_fails(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_application_data,
    dummy_application_dir,
    dummy_domain,
    cli_runner,
    mocker,
):
    response_data = dummy_application_data[0]
    application_id = response_data["id"]

    create_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-script-templates")
    create_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=response_data,
        ),
    )

    upload_route = respx_mock.put(f"{dummy_domain}/jobbergate/job-script-templates/{application_id}")
    upload_route.mock(
        return_value=httpx.Response(httpx.codes.BAD_REQUEST),
    )

    test_app = make_test_app("create", create)
    mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.render_single_result")
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                create --name=dummy-name --identifier=dummy-identifier
                       --application-path={dummy_application_dir}
                       --application-desc="This application is kinda dumb, actually"
                """
            )
        ),
    )
    assert result.exit_code == 1, f"create failed: {result.output}"
    assert create_route.called
    assert upload_route.called
    assert "application files could not be uploaded" in result.output

    mocked_render.assert_called_once_with(
        dummy_context,
        {**response_data, "application_uploaded": False},
        title="Created Application",
        hidden_fields=HIDDEN_FIELDS,
    )


@pytest.mark.parametrize(
    "selector_template",
    [
        "{id}",
        "-i {id}",
        "--id={id}",
        "--id {id}",
        "{identifier}",
        "--identifier={identifier}",
        "--identifier {identifier}",
    ],
)
def test_update__success_by_id(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_application_data,
    dummy_application_dir,
    dummy_domain,
    cli_runner,
    mocker,
    selector_template,
):
    application_data = dummy_application_data[0]
    id = application_data["id"]
    identifier = application_data["identifier"]

    url_selector = identifier if "identifier" in selector_template else id
    cli_selector = selector_template.format(id=id, identifier=identifier)

    update_route = respx_mock.put(f"{dummy_domain}/jobbergate/job-script-templates/{url_selector}")
    update_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=application_data,
        ),
    )

    mocked_upload = mocker.patch("jobbergate_cli.subapps.applications.app.upload_application")
    mocked_upload.return_value = True

    get_route = respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates/{url_selector}")
    get_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=application_data,
        ),
    )

    test_app = make_test_app("update", update)
    mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.render_single_result")
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                update {cli_selector} --update-identifier=dummy-identifier
                       --application-path={dummy_application_dir}
                       --application-desc="This application is kinda dumb, actually"
                """
            )
        ),
    )
    assert result.exit_code == 0, f"update failed: {result.output}"
    assert update_route.called
    assert get_route.called
    assert mocked_upload.called

    mocked_render.assert_called_once_with(
        dummy_context,
        ApplicationResponse(**application_data),
        title="Updated Application",
        hidden_fields=HIDDEN_FIELDS,
    )


def test_update__does_not_upload_if_application_path_is_not_supplied(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_application_data,
    dummy_domain,
    cli_runner,
    mocker,
):
    response_data = dummy_application_data[0]
    application_id = response_data["id"]

    update_route = respx_mock.put(f"{dummy_domain}/jobbergate/job-script-templates/{application_id}")
    update_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=response_data,
        ),
    )

    get_route = respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates/{application_id}")
    get_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=response_data,
        ),
    )

    mocked_upload = mocker.patch("jobbergate_cli.subapps.applications.app.upload_application")

    test_app = make_test_app("update", update)
    mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.render_single_result")
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                update --id={application_id} --update-identifier=dummy-identifier
                       --application-desc="This application is kinda dumb, actually"
                """
            )
        ),
    )
    assert result.exit_code == 0, f"update failed: {result.output}"
    assert update_route.called
    assert mocked_upload.call_count == 0

    mocked_render.assert_called_once_with(
        dummy_context,
        ApplicationResponse(**response_data),
        title="Updated Application",
        hidden_fields=HIDDEN_FIELDS,
    )


def test_update__warns_but_does_not_abort_if_upload_fails(
    respx_mock,
    make_test_app,
    dummy_context,
    dummy_application_data,
    dummy_application_dir,
    dummy_domain,
    cli_runner,
    mocker,
):
    response_data = dummy_application_data[0]
    application_id = response_data["id"]

    update_route = respx_mock.put(f"{dummy_domain}/jobbergate/job-script-templates/{application_id}")
    update_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=response_data,
        ),
    )

    get_route = respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates/{application_id}")
    get_route.mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=response_data,
        ),
    )
    mocked_upload = mocker.patch(
        "jobbergate_cli.subapps.applications.app.upload_application",
        side_effect=Abort("Failed to upload application files"),
    )

    test_app = make_test_app("update", update)
    mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.render_single_result")
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            unwrap(
                f"""
                update --id={application_id} --update-identifier=dummy-identifier
                       --application-path={dummy_application_dir}
                       --application-desc="This application is kinda dumb, actually"
                """
            )
        ),
    )
    assert result.exit_code == 1, f"update failed: {result.output}"
    assert update_route.called
    assert mocked_upload.called
    assert "application files could not be uploaded" in result.output

    mocked_render.assert_called_once_with(
        dummy_context,
        ApplicationResponse(**response_data),
        title="Updated Application",
        hidden_fields=HIDDEN_FIELDS,
    )


@pytest.mark.parametrize(
    "selector_template",
    [
        "{id}",
        "-i {id}",
        "--id={id}",
        "--id {id}",
        "{identifier}",
        "--identifier={identifier}",
        "--identifier {identifier}",
    ],
)
def test_delete__success(respx_mock, make_test_app, dummy_domain, cli_runner, selector_template):
    id = 1
    identifier = "some-identifier"

    url_selector = identifier if "identifier" in selector_template else id
    cli_selector = selector_template.format(id=id, identifier=identifier)

    delete_route = respx_mock.delete(f"{dummy_domain}/jobbergate/job-script-templates/{url_selector}")
    delete_route.mock(return_value=httpx.Response(httpx.codes.NO_CONTENT))

    test_app = make_test_app("delete", delete)
    result = cli_runner.invoke(test_app, shlex.split(f"delete {cli_selector}"))
    assert result.exit_code == 0, f"delete failed: {result.output}"
    assert delete_route.called
    assert "Application delete succeeded" in result.output


class TestDownloadApplicationFiles:
    """
    Test the download application files subcommand.
    """

    @pytest.fixture()
    def test_app(self, make_test_app):
        """
        Fixture to create a test app.
        """
        return make_test_app("download", download_files)

    @pytest.mark.parametrize(
        "selector_template",
        [
            "{id}",
            "-i {id}",
            "--id={id}",
            "--id {id}",
            "{identifier}",
            "--identifier={identifier}",
            "--identifier {identifier}",
        ],
    )
    def test_download__success(
        self,
        respx_mock,
        test_app,
        dummy_application_data,
        dummy_domain,
        dummy_context,
        cli_runner,
        mocker,
        tmp_path,
        selector_template,
    ):
        """
        Test that the download application files subcommand works as expected.
        """
        application_data = dummy_application_data[0]
        id = application_data["id"]
        identifier = application_data["identifier"]

        url_selector = identifier if "identifier" in selector_template else id
        cli_selector = selector_template.format(id=id, identifier=identifier)

        respx_mock.get(f"{dummy_domain}/jobbergate/job-script-templates/{url_selector}").mock(
            return_value=httpx.Response(
                httpx.codes.OK,
                json=application_data,
            ),
        )
        mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.terminal_message")

        list_of_files = [f"file-{i}" for i in range(3)]

        mocked_save_files = mocker.patch(
            "jobbergate_cli.subapps.applications.app.save_application_files", return_value=list_of_files
        )

        with mock.patch.object(pathlib.Path, "cwd", return_value=tmp_path):
            result = cli_runner.invoke(test_app, shlex.split(f"download {cli_selector}"))

        mocked_save_files.assert_called_with(
            dummy_context,
            application_data=ApplicationResponse(**application_data),
            destination_path=tmp_path,
        )

        assert result.exit_code == 0, f"download failed: {result.output}"
        mocked_render.assert_called_once_with(
            f"A total of {len(list_of_files)} application files were successfully downloaded.",
            subject="Application download succeeded",
        )

    def test_download__fails_with_neither_id_or_identifier(self, test_app, cli_runner):
        """
        Test that the download application files subcommand fails when neither id nor identifier are supplied.
        """
        result = cli_runner.invoke(test_app, shlex.split("download"))
        assert result.exit_code != 0
        assert "You must supply one and only one selection value" in result.output

    def test_download__fails_with_both_id_and_identifier(self, test_app, cli_runner):
        """
        Test that the download application files subcommand fails when both id and identifier are supplied.
        """
        result = cli_runner.invoke(test_app, shlex.split("download --id=1 --identifier=dummy"))
        assert result.exit_code != 0
        assert "You must supply one and only one selection value" in result.output


@pytest.mark.parametrize(
    "selector_template",
    [
        "{id}",
        "-i {id}",
        "--id={id}",
        "--id {id}",
        "{identifier}",
        "--identifier={identifier}",
        "--identifier {identifier}",
    ],
)
def test_clone__success(
    respx_mock,
    make_test_app,
    dummy_application_data,
    dummy_domain,
    dummy_context,
    cli_runner,
    mocker,
    selector_template,
):
    """
    Test that the clone application subcommand works as expected.
    """

    application_data = dummy_application_data[0]
    id = application_data["id"]
    identifier = application_data["identifier"]

    url_selector = identifier if "identifier" in selector_template else id
    cli_selector = selector_template.format(id=id, identifier=identifier)

    clone_route = respx_mock.post(f"{dummy_domain}/jobbergate/job-script-templates/clone/{url_selector}").mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=application_data,
        ),
    )
    mocked_render = mocker.patch("jobbergate_cli.subapps.applications.app.render_single_result")

    test_app = make_test_app("clone", clone)
    result = cli_runner.invoke(
        test_app,
        shlex.split(
            "clone {} --application-identifier={} --application-name={} --application-desc={}".format(
                cli_selector,
                shlex.quote(application_data["identifier"]),
                shlex.quote(application_data["name"]),
                shlex.quote(application_data["description"]),
            ),
        ),
    )

    assert clone_route.called

    assert result.exit_code == 0, f"clone failed: {result.output}"
    mocked_render.assert_called_once_with(
        dummy_context,
        ApplicationResponse(**application_data),
        title="Cloned Application",
        hidden_fields=HIDDEN_FIELDS,
    )



import contextlib
import os
import pathlib

import pytest


@pytest.fixture
def temp_cd():
    @contextlib.contextmanager
    def _helper(path: pathlib.Path):
        """
        Helper method to temporarily change directory.
        Pretty much copied from:
        https://dev.to/teckert/changing-directory-with-a-python-context-manager-2bj8
        """
        anchor = pathlib.Path().absolute()
        try:
            os.chdir(path)
            yield
        finally:
            os.chdir(anchor)

    return _helper
